import collections
from collections.abc import Sequence
from typing import Any, NamedTuple

import peewee as pw
from playhouse.reflection import Column as ColumnSerializer

from miggy.utils import ModelIndex, get_default_constraint, get_default_constraint_value, indexes_state

from .types import ModelCls

INDENT = "    "
NEWLINE = "\n" + INDENT
FIELD_MODULES_MAP = {
    "ArrayField": "pw_pext",
    "BinaryJSONField": "pw_pext",
    "DateTimeTZField": "pw_pext",
    "HStoreField": "pw_pext",
    "IntervalField": "pw_pext",
    "JSONField": "pw_pext",
    "TSVectorField": "pw_pext",
}


def fk_to_params(field: pw.ForeignKeyField) -> dict[str, Any]:
    params = {"model": field.rel_model._meta.name}
    if field.on_delete is not None:
        params["on_delete"] = "'%s'" % field.on_delete
    if field.on_update is not None:
        params["on_update"] = "'%s'" % field.on_update
    if field.constraint_name is not None:
        params["constraint_name"] = "'%s'" % field.constraint_name
    return params


TYPE_PARAMS = {
    pw.CharField: lambda f: {"max_length": f.max_length},
    pw.DecimalField: lambda f: {
        "max_digits": f.max_digits,
        "decimal_places": f.decimal_places,
    },
}


def get_type_params(field: pw.Field) -> dict[str, Any]:
    field_type = type(field)
    params = TYPE_PARAMS.get(field_type, lambda f: {})(field)
    return params


def get_field_params(field: pw.Field) -> dict[str, Any]:
    params = get_type_params(field)
    if isinstance(field, pw.ForeignKeyField):
        params.update(fk_to_params(field))
    return params


def field_to_code(field, space=True) -> str:
    serializer = FieldSerializer(field)
    return serializer.serialize(" " if space else "")


def _get_default(field: pw.Field) -> Any:
    if field.default is not None and not callable(field.default):
        return field.default
    return None


def field_to_params(field: pw.Field) -> dict[str, Any]:
    params = get_field_params(field)
    params["type"] = type(field)
    params["null"] = field.null
    params["column_name"] = field.column_name
    params["default"] = _get_default(field)
    params["default_constraint"] = get_default_constraint_value(field)
    params["index"] = field.index and not field.unique, field.unique
    return params


def fields_not_equal(field1: pw.Field, field2: pw.Field) -> bool:
    return field_to_params(field1) != field_to_params(field2)


class FieldSerializer(ColumnSerializer):
    def __init__(self, field: pw.Field) -> None:
        self.field = field
        super(FieldSerializer, self).__init__(
            field.name,
            type(field),
            field.field_type,
            field.null,
            primary_key=field.primary_key,
            column_name=field.column_name,
            index=field.index,
            unique=field.unique,
            extra_parameters={},
        )
        if field.default is not None and not callable(field.default):
            self.default = repr(field.default)

        self.extra_parameters.update(get_field_params(field))

        self.rel_model = None
        self.related_name = None
        self.to_field = None

        if isinstance(field, pw.ForeignKeyField):
            self.to_field = field.rel_field.name
            self.related_name = field.backref
            self.rel_model = "migrator.orm['%s']" % field.rel_model._meta.name

    def get_field_parameters(self) -> dict[str, Any]:
        params = super(FieldSerializer, self).get_field_parameters()
        # original method put value from default in constraints so override this logic
        params.pop("constraints", None)
        if self.field.constraints:
            default_constraint = get_default_constraint(self.field)
            if default_constraint is not None:
                params["constraints"] = '[pw.SQL("DEFAULT %s")]' % default_constraint.value.replace('"', '\\"')
        if self.default is not None:
            params["default"] = self.default
        return params

    def serialize(self, space=" ") -> str:
        # Generate the field definition for this column.
        field = self.get_field()
        module = FIELD_MODULES_MAP.get(self.field_class.__name__, "pw")
        name, _, field = [s and s.strip() for s in field.partition("=")]
        return "{name}{space}={space}{module}.{field}".format(name=name, field=field, space=space, module=module)


class IndexMeta(NamedTuple):
    model: str
    fields: tuple[str, ...]
    name: str
    unique: bool = False
    where: str | None = None


def resolve_field(model_cls: pw.Model, field: str) -> pw.Field:
    _field = model_cls._meta.combined.get(field, None)
    if _field is None:
        raise ValueError(f"{model_cls} does not have '{field}' field.")
    return _field


class IndexMetaExtractor:
    def __init__(self, model_cls: ModelCls, index_obj: ModelIndex) -> None:
        self.model_cls = model_cls
        self.index_obj = index_obj

    def resolve_where(self, where: pw.Node | None) -> None | str:
        if isinstance(where, pw.SQL):
            if where.params is not None:
                raise NotImplementedError(
                    "SQL object with params for where condition is not suported. Use SQL object without params instead."
                )
            return where.sql
        if where is None:
            return None
        raise NotImplementedError(
            f"{type(where)} for where condition is not suported. Use SQL object without params instead."
        )

    def validate_fields(self, model_index: pw.ModelIndex) -> None:
        for f in model_index._expressions:
            if not isinstance(f, pw.Field):
                raise NotImplementedError(f"{type(f)} for ModelIndex.field is not suported. Use Field object instead.")

    def serialize(self) -> IndexMeta:
        model_index = self.index_obj
        self.validate_fields(model_index)
        return IndexMeta(
            self.model_cls._meta.name,
            tuple(f.name for f in model_index._expressions),
            unique=model_index._unique,
            where=self.resolve_where(model_index._where),
            name=model_index._name,
        )


def extract_index_meta(model_cls: ModelCls) -> list[IndexMeta]:
    rebuild_indexes(model_cls)
    indexes = indexes_state(model_cls).values()
    return [IndexMetaExtractor(model_cls, i).serialize() for i in indexes]


def rebuild_indexes(model_cls: ModelCls) -> None:
    def resolve_fields(fields: Sequence[Any]) -> tuple[str, ...]:
        _fields = []
        for field in fields:
            if isinstance(field, str):
                field = resolve_field(model_cls, field)
            if isinstance(field, pw.Field):
                _fields.append(field)
            else:
                raise NotImplementedError
        return tuple(_fields)

    for index_obj in model_cls._meta.indexes:
        # Advanced Indexes
        # https://docs.peewee-orm.com/en/latest/peewee/models.html#advanced-index-creation
        if isinstance(index_obj, pw.ModelIndex):
            indexes_state(model_cls)[index_obj._name] = index_obj

        # Multi-column indexes
        # https://docs.peewee-orm.com/en/latest/peewee/models.html#multi-column-indexes
        elif isinstance(index_obj, (list, tuple)):
            fields, unique = index_obj
            model_index = ModelIndex(model_cls, resolve_fields(fields), unique=unique)
            indexes_state(model_cls)[model_index._name] = model_index
        else:
            raise NotImplementedError(
                f"{type(index_obj)} as Index is not suported. Use ModelIndex, list or tuple instead."
            )
    model_cls._meta.indexes = []


def diff_indexes_from_meta(current: ModelCls, prev: ModelCls) -> tuple[list[str], list[str]]:
    create_changes = []
    drop_changes = []
    current_indexes = extract_index_meta(current)
    prev_indexes = extract_index_meta(prev)

    for index in set(current_indexes) - set(prev_indexes):
        create_changes.append(add_index(index))
    for index in set(prev_indexes) - set(current_indexes):
        drop_changes.append(drop_index(prev, index.name))
    return create_changes, drop_changes


def diff_one(current: ModelCls, prev: ModelCls) -> list[str]:
    """Find difference between given peewee models."""
    changes = []

    fields1 = current._meta.fields
    fields2 = prev._meta.fields

    if current._meta.table_name != prev._meta.table_name:
        changes.append(rename_table(prev, current._meta.table_name))

    create_index_changes, drop_index_changes = diff_indexes_from_meta(current, prev)

    # Drop non-field indexes before dropping and creating fields
    changes.extend(drop_index_changes)

    # Add fields
    names1 = set(fields1) - set(fields2)
    if names1:
        fields = [fields1[name] for name in names1]
        changes.append(add_fields(current, *fields))

    # Drop fields
    names2 = set(fields2) - set(fields1)
    if names2:
        changes.append(remove_fields(current, *names2))

    # Change fields
    fields_ = []
    for name in set(fields1) - names1 - names2:
        field1, field2 = fields1[name], fields2[name]
        if fields_not_equal(field1, field2):
            fields_.append(field1)

    if fields_:
        changes.append(change_fields(current, *fields_))

    # Create non-field indexes after dropping and creating fields
    changes.extend(create_index_changes)

    return changes


def diff_many(current_models, prev_models, reverse=False):
    """Calculate changes for migrations from models2 to models1."""
    if reverse:
        prev_models, current_models = current_models, prev_models
    current_models = pw.sort_models(current_models)
    prev_models = pw.sort_models(prev_models)

    if reverse:
        current_models = reversed(current_models)
        prev_models = reversed(prev_models)

    current_models = collections.OrderedDict([(m._meta.name, m) for m in current_models])
    prev_models = collections.OrderedDict([(m._meta.name, m) for m in prev_models])

    changes = []

    for name, model1 in current_models.items():
        if name not in prev_models:
            continue
        changes += diff_one(model1, prev_models[name])

    # Add models
    for name in [m for m in current_models if m not in prev_models]:
        index_meta = extract_index_meta(current_models[name])
        changes.append(create_model(current_models[name]))
        for i in index_meta:
            changes.append(add_index(i))

    # Remove models
    for name in [m for m in prev_models if m not in current_models]:
        changes.append(remove_model(prev_models[name]))

    return changes


def model_to_code(Model) -> str:
    template = """class {classname}(pw.Model):
{fields}

{meta}
"""
    fields = INDENT + NEWLINE.join(
        [
            field_to_code(field)
            for field in Model._meta.sorted_fields
            if not (isinstance(field, pw.PrimaryKeyField) and field.name == "id")
        ]
    )
    meta = INDENT + NEWLINE.join(
        filter(
            None,
            [
                "class Meta:",
                INDENT + 'table_name = "%s"' % Model._meta.table_name,
                (INDENT + 'schema = "%s"' % Model._meta.schema) if Model._meta.schema else "",
                (INDENT + "primary_key = pw.CompositeKey{0}".format(Model._meta.primary_key.field_names))
                if isinstance(Model._meta.primary_key, pw.CompositeKey)
                else "",
            ],
        )
    )

    return template.format(classname=Model.__name__, fields=fields, meta=meta)


def create_model(model: ModelCls) -> str:
    return "@migrator.create_model\n" + model_to_code(model)


def remove_model(model: ModelCls) -> str:
    return "migrator.remove_model('%s')" % model._meta.name


def add_fields(model: ModelCls, *fields) -> str:
    return "migrator.add_fields(%s'%s', %s)" % (
        NEWLINE,
        model._meta.name,
        NEWLINE + ("," + NEWLINE).join([field_to_code(field, False) for field in fields]),
    )


def remove_fields(model: ModelCls, *fields) -> str:
    return "migrator.remove_fields('%s', %s)" % (model._meta.name, ", ".join(map(repr, fields)))


def change_fields(model: ModelCls, *fields) -> str:
    return "migrator.change_fields('%s', %s)" % (
        model._meta.name,
        ("," + NEWLINE).join([field_to_code(f, False) for f in fields]),
    )


def change_not_null(model: ModelCls, name, null) -> str:
    operation = "drop_not_null" if null else "add_not_null"
    return "migrator.%s('%s', %s)" % (operation, model._meta.name, repr(name))


def add_index(index_meta: IndexMeta) -> str:
    operation = "add_index"

    _where = ", where=pw.SQL(%s)" % repr(index_meta.where) if index_meta.where else ""
    _unique = f", unique={index_meta.unique}" if index_meta.unique else ""
    _fields = ", ".join(map(repr, index_meta.fields))
    return f"migrator.{operation}('{index_meta.model}', {_fields}, name='{index_meta.name}'{_unique}{_where})"


def drop_index(model: ModelCls, name: str) -> str:
    operation = "drop_index"
    return "migrator.%s('%s', '%s')" % (operation, model._meta.name, name)


def rename_table(model: ModelCls, new_name: str) -> str:
    operation = "rename_table"
    return "migrator.%s('%s', '%s')" % (operation, model._meta.name, new_name)



from collections.abc import Callable, ValuesView
from typing import Any

import peewee as pw
from playhouse.migrate import (
    SQL,
    MySQLDatabase,
    Operation,
    PostgresqlDatabase,
    SqliteDatabase,
    operation,
)
from playhouse.migrate import MySQLMigrator as MqM
from playhouse.migrate import PostgresqlMigrator as PgM
from playhouse.migrate import SchemaMigrator as ScM
from playhouse.migrate import SqliteMigrator as SqM

from miggy import LOGGER
from miggy.auto import fk_to_params, get_type_params, resolve_field
from miggy.types import ModelCls
from miggy.utils import (
    ModelIndex,
    copy_model,
    delete_field,
    get_default_constraint,
    get_default_constraint_value,
    get_single_index,
    get_single_index_name,
    has_single_index,
    indexes_state,
    make_single_index,
)

ModelDict = dict[str, ModelCls]
RunPythonF = Callable[["SchemaMigrator", "State"], None]


class State:
    """
    Current state containing historical models that match the operation’s place in the project history.
    This is a dict-like class that stores data in the format model_name: model_class.
    The model_name is case-insensitive.

    Example::

        User = state["user"]
        User.get(id=1)
    """

    def __init__(self, data: ModelDict | None = None) -> None:
        self.data: ModelDict = data or {}
        self._snapshot: ModelDict | None = None

    def normalize_key(self, key: str) -> str:
        _key = key.lower()
        if self._snapshot is not None:
            if _key in self._snapshot:
                self._snapshot[_key] = copy_model(self._snapshot[_key])
        return _key

    def __setitem__(self, key: str, val: ModelCls) -> None:
        self.data[self.normalize_key(key)] = val

    def __getitem__(self, key: str) -> ModelCls:
        return self.data[self.normalize_key(key)]

    def __delitem__(self, key: str) -> None:
        del self.data[self.normalize_key(key)]

    def __contains__(self, key: str) -> bool:
        return self.normalize_key(key) in self.data

    def values(self) -> ValuesView[ModelCls]:
        return self.data.values()

    def create_snapshot(self) -> None:
        self._snapshot = self.data.copy()

    def pop_snapshot(self) -> "State":
        _snapshot = self._snapshot
        self._snapshot = None
        return State(_snapshot)


class MigrateOperation:
    """
    Base class for a migrate operation
    """

    def state_forwards(self, state: State) -> None:
        """
        Take the state from the previous migration, and mutate it
        so that it matches what this migration would perform.
        """

        raise NotImplementedError

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation] | list[Callable]:
        """
        Perform the mutation on the database schema in the normal
        (forwards) direction.
        The method MUST NOT mutate provided states.
        """
        raise NotImplementedError


class RunPython(MigrateOperation):
    """
    Allows to run custom Python code. **func** should be callable object that accept two arguments;
    the first is an instance of :class:`SchemaMigrator` and the second  is an instance of :class:`State`

    Example::

        def save_user(schema_migrator: SchemaMigrator, current_state: State):
            User = current_state["user"]
            User(
                first_name="First",
                last_name="Last",
            ).save()

        migrator.add_operaion(RunPython(save_user))
    """

    def __init__(self, func: RunPythonF) -> None:
        self.func = func

    def state_forwards(self, state: State) -> None:
        pass

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation] | list[Callable]:
        return [lambda: self.func(schema_migrator, from_state)]


class RunSql(MigrateOperation):
    """
    Allows running of arbitrary SQL on the database -
    useful for more advanced features of database backends that Miggy doesn’t support directly.

    Example::

        migrator.add_operation(
            RunSql(
                'INSERT INTO "user" ("first_name", "last_name") VALUES (%s, %s)',
                (
                    "First",
                    "Last",
                ),
            )
        )
    """

    def __init__(self, sql: str, params: tuple[Any, ...] | None = None) -> None:
        self.sql = sql
        self.params = params

    def state_forwards(self, state: State) -> None:
        pass

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation] | list[Callable]:
        return [schema_migrator.sql(self.sql, self.params)]


class CreateModel(MigrateOperation):
    """
    Creates a new model in the :class:`State` and a corresponding table in the database to match it.
    """

    def __init__(self, model: ModelCls) -> None:
        self.model = model

    def state_forwards(self, state: State) -> None:
        state[self.model._meta.name] = self.model

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Callable]:
        model = to_state[self.model._meta.name]
        return [schema_migrator.create_table(model)]


class RemoveModel(MigrateOperation):
    """
    Deletes the model from the :class:`State` and its table from the database.
    """

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name

    def state_forwards(self, state: State) -> None:
        del state[self.model_name]

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Callable]:
        model = from_state[self.model_name]
        return [schema_migrator.drop_table(model)]


class AddIndex(MigrateOperation):
    """
    Creates an index in the database table for the model with model_name.
    The index will be saved in **Model._meta.indexes_state** dict
    """

    def __init__(
        self,
        model_name: str,
        *fields: str,
        name: str,
        unique: bool = False,
        where: pw.SQL | None = None,
        safe: bool = False,
        concurrently: bool = False,
    ) -> None:
        self.model_name = model_name
        self.fields = fields
        self.unique = unique
        self.where = where
        self.name = name
        self.safe = safe
        self.concurrently = concurrently
        self._index: ModelIndex | None = None

    def build_index(self, model: ModelCls) -> ModelIndex:
        if not self._index:
            self._index = ModelIndex(
                model=model,
                fields=[resolve_field(model, f) for f in self.fields],
                unique=self.unique,
                where=self.where,
                name=self.name,
                safe=self.safe,
                concurrently=self.concurrently,
            )
        return self._index

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        model_index = self.build_index(model)
        indexes_state(model)[model_index._name] = model_index

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        model = to_state[self.model_name]
        model_index = self.build_index(model)
        return [schema_migrator.add_model_index(model_index)]


class DropIndex(MigrateOperation):
    """
    Removes the index named name from the model with model_name.
    """

    def __init__(self, model_name: str, name: str) -> None:
        self.model_name = model_name
        self.name = name

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        del indexes_state(model)[self.name]

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        model = from_state[self.model_name]
        return [schema_migrator.drop_index(model._meta.table_name, self.name)]


class RenameTable(MigrateOperation):
    """
    Renames the model from the old name to a new one.
    It also renames all single-column indexes, if they exist.

    **Warning:**

    This operation does not rename indexes created via the **Meta** class or the **add_index()** method.
    You should explicitly specify index names if you plan to use this operation.
    Otherwise, you will be prompted to recreate the indexes with a new name in the next migration.
    """

    def __init__(self, model_name: str, new_table_name: str) -> None:
        self.model_name = model_name
        self.new_table_name = new_table_name

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        model._meta.table_name = self.new_table_name

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Callable]:
        """Rename table in database."""
        old_model = from_state[self.model_name]
        new_model = to_state[self.model_name]
        ops = [schema_migrator.rename_table(old_model._meta.table_name, self.new_table_name)]
        for old_field in old_model._meta.sorted_fields:
            new_field = getattr(new_model, old_field.name)
            ops.append(schema_migrator.resolve_single_index_name(old_field, new_field))
        return ops


class AddFields(MigrateOperation):
    """
    Adds fields to a model.
    """

    def __init__(self, model_name: str, **fields: pw.Field) -> None:
        self.model_name = model_name
        self.fields = fields

    def state_forwards(self, state: State) -> None:
        for name, field in self.fields.items():
            state[self.model_name]._meta.add_field(name, field)

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        ops = []
        model = to_state[self.model_name]
        for field in self.fields.values():
            ops.append(schema_migrator.add_column(model._meta.table_name, field.column_name, field))
        return ops


class ChangeFields(MigrateOperation):
    """
    Change fields to a model.
    """

    def __init__(self, model_name: str, **fields: pw.Field) -> None:
        self.model_name = model_name
        self.fields = fields

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        for name, field in self.fields.items():
            model._meta.add_field(name, field)

    def handle_indexes(
        self, old_field: pw.Field, new_field: pw.Field, schema_migrator: "SchemaMigrator"
    ) -> list[Operation]:
        _ops = []
        _field = new_field
        if _field.unique and old_field.unique:
            return []
        if not _field.unique and not old_field.unique and _field.index == old_field.index:
            return []
        table_name = old_field.model._meta.table_name
        if has_single_index(old_field):
            # We have already renamed the column so create name from the new field
            _ops.append(schema_migrator.drop_index(table_name, get_single_index_name(_field)))
        if model_index := get_single_index(_field):
            _ops.append(schema_migrator.add_model_index(model_index))
        return _ops

    def handle_fk_constraint(
        self, old_field: pw.Field, new_field: pw.Field, schema_migrator: "SchemaMigrator"
    ) -> list[Operation]:
        _ops: list[Operation] = []
        is_old_field_fk = isinstance(old_field, pw.ForeignKeyField)
        is_new_field_fk = isinstance(new_field, pw.ForeignKeyField)
        if is_old_field_fk and is_new_field_fk and fk_to_params(old_field) == fk_to_params(new_field):
            # Nothing's changed for fk
            return _ops
        table_name = old_field.model._meta.table_name
        if is_old_field_fk:
            _ops.append(schema_migrator.drop_foreign_key_constraint(table_name, new_field.column_name))
        if is_new_field_fk:
            _ops.append(
                schema_migrator.add_foreign_key_constraint(
                    table_name,
                    new_field.column_name,
                    new_field.rel_model._meta.table_name,
                    new_field.rel_field.name,
                    new_field.on_delete,
                    new_field.on_update,
                    constraint_name=new_field.constraint_name,
                )
            )
        return _ops

    def handle_default_constraint(
        self, old_field: pw.Field, new_field: pw.Field, schema_migrator: "SchemaMigrator"
    ) -> list[Operation]:
        old_value = get_default_constraint_value(old_field) or ""
        new_value = get_default_constraint_value(new_field) or ""
        table_name = old_field.model._meta.table_name
        if old_value != new_value:
            if new_value:
                return [schema_migrator.add_column_default(table_name, new_field.column_name, new_value)]
            else:
                return [
                    schema_migrator.drop_column_default(
                        table_name,
                        new_field.column_name,
                    )
                ]
        return []

    def handle_type(
        self, old_field: pw.Field, new_field: pw.Field, schema_migrator: "SchemaMigrator"
    ) -> list[Operation]:
        if type(old_field) is not type(new_field) or get_type_params(old_field) != get_type_params(new_field):
            table_name = old_field.model._meta.table_name
            return [schema_migrator.alter_column_type(table_name, new_field.column_name, new_field)]
        return []

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        _ops = []
        model = from_state[self.model_name]
        table_name = model._meta.table_name
        for name, field in self.fields.items():
            old_field = getattr(model, name)
            old_column_name = old_field.column_name

            if old_column_name != field.column_name:
                _ops.append(schema_migrator.rename_field(table_name, old_field, field))

            _ops.extend(self.handle_type(old_field, field, schema_migrator))
            _ops.extend(self.handle_fk_constraint(old_field, field, schema_migrator))
            _ops.extend(self.handle_default_constraint(old_field, field, schema_migrator))
            if old_field.null != field.null:
                _operation = schema_migrator.drop_not_null if field.null else schema_migrator.add_not_null
                _ops.append(_operation(table_name, field.column_name))
            _ops.extend(self.handle_indexes(old_field, field, schema_migrator))
        return _ops


class RemoveFields(MigrateOperation):
    """
    Removes fields from a model
    """

    def __init__(self, model_name: str, *names: str, cascade: bool = False) -> None:
        self.model_name = model_name
        self.cascade = cascade
        self.names = names

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        for name in self.names:
            field = state[self.model_name]._meta.fields[name]
            delete_field(model, field)

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        """Remove fields from model."""
        ops = []
        model = from_state[self.model_name]
        for name in self.names:
            field = model._meta.fields[name]
            ops.append(schema_migrator.drop_column(model._meta.table_name, field.column_name, cascade=self.cascade))
        return ops


def fk_postfix(name: str) -> str:
    return name if name.endswith("_id") else name + "_id"


class RenameField(MigrateOperation):
    """
    Changes a field’s name (and, unless **column_name** is set, its column name).
    It also renames a single-column indexe, if it exists.

    **Warning:**

    This operation does not rename indexes created via the **Meta** class or the **add_index()** method.
    You should explicitly specify index names if you plan to use this operation.
    Otherwise, you will be prompted to recreate the index with a new name in the next migration.
    """

    def __init__(self, model_name: str, old_name: str, new_name: str) -> None:
        self.model_name = model_name
        self.old_field_name = old_name
        self.new_field_name = new_name

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]

        old_field = model._meta.fields[self.old_field_name]
        new_field = old_field.clone()
        delete_field(model, old_field)

        new_field.column_name = self.resolve_new_name(old_field, self.new_field_name)
        model._meta.add_field(self.new_field_name, new_field)

    def resolve_new_name(self, old_field: pw.Field, new_name: str) -> str:
        if isinstance(old_field, pw.ForeignKeyField):
            if old_field.column_name == fk_postfix(old_field.name):
                return fk_postfix(new_name)
        if old_field.column_name == old_field.name:
            return new_name
        return old_field.column_name

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        old_model = from_state[self.model_name]
        new_model = to_state[self.model_name]
        old_field = old_model._meta.fields[self.old_field_name]
        new_field = new_model._meta.fields[self.new_field_name]
        if old_field.column_name != new_field.column_name:
            return [schema_migrator.rename_field(new_model._meta.table_name, old_field, new_field)]
        return []


class ChangeNullable(MigrateOperation):
    def __init__(self, model_name: str, *names: str, is_null: bool) -> None:
        self.model_name = model_name
        self.names = names
        self.is_null = is_null

    def state_forwards(self, state: State) -> None:
        model = state[self.model_name]
        for name in self.names:
            field = model._meta.fields[name]
            field.null = self.is_null

    def database_forwards(
        self, schema_migrator: "SchemaMigrator", from_state: State, to_state: State
    ) -> list[Operation]:
        ops = []
        model = to_state[self.model_name]
        for name in self.names:
            field = model._meta.fields[name]
            _operation = schema_migrator.drop_not_null if self.is_null else schema_migrator.add_not_null
            ops.append(_operation(model._meta.table_name, field.column_name))
        return ops


class Migration:
    def __init__(self, state: State, schema_migrator: "SchemaMigrator", schema: str | None = None) -> None:
        self.state = state
        self.schema_migrator = schema_migrator
        self.schema = schema
        self.operations: list[Operation | Callable] = []

    def append(self, op: MigrateOperation) -> None:
        self.state.create_snapshot()
        op.state_forwards(self.state)
        from_state = self.state.pop_snapshot()
        self.operations.extend(op.database_forwards(self.schema_migrator, from_state, self.state))

    def apply(self, change_schema: bool) -> None:
        if not change_schema:
            return

        if self.schema:
            _ops = [self.schema_migrator.select_schema(self.schema), *self.operations]
        else:
            _ops = [*self.operations]

        for op in _ops:
            if isinstance(op, Operation):
                LOGGER.info("%s %s", op.method, op.args)
                op.run()
            else:
                op()

    def clean(self) -> None:
        self.operations = []


class SchemaMigrator(ScM):
    """Extended **playhouse.migrate.SchemaMigrator** from **peewee**"""

    @classmethod
    def from_database(cls, database):
        """Initialize migrator by db."""
        if isinstance(database, PostgresqlDatabase):
            return PostgresqlMigrator(database)
        if isinstance(database, SqliteDatabase):
            return SqliteMigrator(database)
        if isinstance(database, MySQLDatabase):
            return MySQLMigrator(database)
        return super(SchemaMigrator, cls).from_database(database)

    @operation
    def select_schema(self, schema):
        """Select database schema"""
        raise NotImplementedError()

    @operation
    def sql(self, sql, params: tuple[Any, ...] | None = None):
        """Execute raw SQL."""
        return SQL(sql, params)

    @operation
    def add_column(self, table, column_name, field):
        # Adding a column is complicated by the fact that if there are rows
        # present and the field is non-null, then we need to first add the
        # column as a nullable field, then set the value, then add a not null
        # constraint.
        default_constraint = get_default_constraint(field)
        if not field.null and field.default is None and not default_constraint:
            raise ValueError("%s is not null but has no default" % column_name)

        is_foreign_key = isinstance(field, pw.ForeignKeyField)
        if is_foreign_key and not field.rel_field:
            raise ValueError("Foreign keys must specify a `field`.")

        operations = [self.alter_add_column(table, column_name, field)]

        # In the event the field is *not* nullable and has no default constraint, update with the default
        # value and set not null.
        if not field.null:
            if not default_constraint:
                operations.append(
                    self.apply_default(table, column_name, field),
                )

            operations.append(self.add_not_null(table, column_name))

        if is_foreign_key and self.explicit_create_foreign_key:
            operations.append(
                self.add_foreign_key_constraint(
                    table,
                    column_name,
                    field.rel_model._meta.table_name,
                    field.rel_field.column_name,
                    field.on_delete,
                    field.on_update,
                )
            )

        if model_index := get_single_index(field):
            operations.append(self.add_model_index(model_index))
        return operations

    @operation
    def add_model_index(self, model_index: ModelIndex):
        ctx = self.make_context()
        return ctx.sql(model_index)

    @operation
    def rename_index(self, old_name: str, new_name: str):
        """Change index name"""
        ctx = self.make_context()
        return ctx.literal("ALTER INDEX ").sql(pw.Entity(old_name)).literal(" RENAME TO ").sql(pw.Entity(new_name))

    @operation
    def resolve_single_index_name(self, old_field: pw.Field, new_field: pw.Field):
        operations = []
        if old_model_index := get_single_index(old_field):
            new_single_index = make_single_index(new_field)
            operations.append(self.rename_index(old_model_index._name, new_single_index._name))
        return operations

    @operation
    def rename_field(self, table: str, old_field: pw.Field, new_field: pw.Field):
        operations = [self.rename_column(table, old_field.column_name, new_field.column_name)]
        operations.append(self.resolve_single_index_name(old_field, new_field))
        return operations

    def create_table(self, model: ModelCls, safe: bool = False) -> Callable:
        """
        Create table from model class
        """
        model._meta.database = self.database
        model._meta.legacy_table_names = False
        return lambda: model.create_table(safe=safe)

    def drop_table(self, model: ModelCls, safe: bool = False) -> Callable:
        """
        Drop model table
        """
        model._meta.database = self.database
        return lambda: model.drop_table(safe=safe)


class MySQLMigrator(SchemaMigrator, MqM):
    def alter_change_column(self, table, column, field):
        """Support change columns."""
        ctx = self.make_context()
        field_null, field.null = field.null, True
        ctx = self._alter_table(ctx, table).literal(" MODIFY COLUMN ").sql(field.ddl(ctx))
        field.null = field_null
        return ctx


class PostgresqlMigrator(SchemaMigrator, PgM):
    """Support the migrations in postgresql."""

    @operation
    def select_schema(self, schema):
        """Select database schema"""
        return self.set_search_path(schema)

    def get_foreign_key_constraint(self, table: str, column_name: str) -> str:
        sql = """
            SELECT DISTINCT
                kcu.constraint_name
            FROM information_schema.table_constraints AS tc
            JOIN information_schema.key_column_usage AS kcu
                ON (tc.constraint_name = kcu.constraint_name AND
                    tc.constraint_schema = kcu.constraint_schema AND
                    tc.table_name = kcu.table_name AND
                    tc.table_schema = kcu.table_schema)
            JOIN information_schema.constraint_column_usage AS ccu
                ON (ccu.constraint_name = tc.constraint_name AND
                    ccu.constraint_schema = tc.constraint_schema)
            WHERE
                tc.constraint_type = 'FOREIGN KEY' AND
                tc.table_name = %s AND
                tc.table_schema = current_schema() AND
                kcu.column_name  = %s"""
        cursor = self.database.execute_sql(sql, (table, column_name))
        return cursor.fetchall()[0][0]

    @operation
    def drop_foreign_key_constraint(self, table: str, column_name: str):
        fk_constraint = self.get_foreign_key_constraint(table, column_name)
        return self.drop_constraint(table, fk_constraint)


class SqliteMigrator(SchemaMigrator, SqM):
    """Support the migrations in sqlite."""

    def drop_table(self, model, cascade=True):
        """SQLite doesnt support cascade syntax by default."""
        return lambda: model.drop_table(cascade=False)

    def alter_column_type(self, table, column, field):
        """Support change columns."""
        return self._update_column(table, column, lambda a, b: b)

    def drop_column(self, table, column_name, cascade=True, legacy=True, **kwargs):
        """drop_column will not work for FK so we should use the legacy version"""
        return super(SqliteMigrator, self).drop_column(table, column_name, cascade, legacy, **kwargs)


class Migrator(object):
    """
    A class that provides shortcuts for adding migration operations.
    """

    def __init__(self, database, schema=None):
        """Initialize the migrator."""
        if isinstance(database, pw.Proxy):
            database = database.obj

        self.database = database
        self.state = State()
        self.schema_migrator = SchemaMigrator.from_database(self.database)
        self.schema = schema

        self.migration = Migration(self.state, self.schema_migrator, schema=schema)

    def add_operation(self, op: MigrateOperation) -> None:
        """
        Adds a migrate operation
        """
        self.migration.append(op)

    def run(self, change_schema: bool = True):
        self.migration.apply(change_schema)
        self.clean()

    def python(self, func: RunPythonF):
        """A shortcut for adding a :class:`RunPython` operation."""
        self.add_operation(RunPython(func))

    def sql(self, sql: str, params: tuple[Any, ...] | None = None) -> None:
        """A shortcut for adding a :class:`RunSql` operation."""
        self.add_operation(RunSql(sql, params))

    def clean(self):
        """Clean the operations."""
        self.migration.clean()

    def create_model(self, model: ModelCls) -> ModelCls:
        """A shortcut for adding a :class:`CreateModel` operation."""

        self.add_operation(CreateModel(model))
        return model

    create_table = create_model

    def remove_model(self, model_name: str) -> None:
        """A shortcut for adding a :class:`RemoveModel` operation."""

        self.add_operation(RemoveModel(model_name))

    drop_table = remove_model

    def add_fields(self, model_name: str, **fields: Any) -> None:
        """A shortcut for adding a :class:`AddFields` operation."""

        self.add_operation(AddFields(model_name, **fields))

    add_columns = add_fields

    def change_fields(self, model_name: str, **fields: pw.Field) -> None:
        """A shortcut for adding a :class:`ChangeFields` operation."""

        return self.add_operation(ChangeFields(model_name, **fields))

    change_columns = change_fields

    def remove_fields(self, model_name: str, *names: str, cascade: bool = False) -> None:
        """A shortcut for adding a :class:`RemoveFields` operation."""

        self.add_operation(RemoveFields(model_name, *names, cascade=cascade))

    drop_columns = remove_fields

    def rename_field(self, model_name: str, old_name: str, new_name: str) -> None:
        """A shortcut for adding a :class:`RenameField` operation."""

        self.add_operation(RenameField(model_name, old_name, new_name))

    rename_column = rename_field

    def rename_table(self, model_name: str, new_table_name: str) -> None:
        """A shortcut for adding a :class:`RenameTable` operation."""

        self.add_operation(RenameTable(model_name, new_table_name))

    rename_model = rename_table

    def add_index(
        self,
        model_name: str,
        *fields: str,
        name: str,
        unique: bool = False,
        where: pw.SQL | None = None,
        safe: bool = False,
        concurrently: bool = False,
    ) -> None:
        """A shortcut for adding a :class:`AddIndex` operation."""

        self.add_operation(
            AddIndex(model_name, *fields, name=name, unique=unique, where=where, safe=safe, concurrently=concurrently)
        )

    def drop_index(self, model_name: str, name: str) -> None:
        """A shortcut for adding a :class:`DropIndex` operation."""

        self.add_operation(DropIndex(model_name, name))

    def add_not_null(self, model_name: str, *names: str) -> None:
        self.add_operation(ChangeNullable(model_name, *names, is_null=False))

    def drop_not_null(self, model_name: str, *names: str) -> None:
        """Drop not null."""
        self.add_operation(ChangeNullable(model_name, *names, is_null=True))



import os
import pkgutil
import re
import sys
import typing
from functools import cached_property
from importlib import import_module

import peewee as pw
from playhouse.psycopg3_ext import Psycopg3Database

from miggy import LOGGER, MigrateHistory
from miggy.auto import NEWLINE, diff_many
from miggy.migrator import Migrator
from miggy.utils import exec_in

CLEAN_RE = re.compile(r"\s+$", re.M)
CURDIR = os.getcwd()
DEFAULT_MIGRATE_DIR = os.path.join(CURDIR, "migrations")
UNDEFINED = object()
VOID = lambda m, d: None  # noqa
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "template.txt")) as t:
    MIGRATE_TEMPLATE = t.read()


class Migration:
    atomic = True

    @staticmethod
    def migrate(migrator, database, fake=False) -> None:
        pass

    @staticmethod
    def rollback(migrator, database, fake=False) -> None:
        pass


class BaseRouter(object):
    """Abstract base class for router."""

    def __init__(self, database, migrate_table="migratehistory", ignore=None, schema=None, logger=LOGGER):
        self.database = database
        self.migrate_table = migrate_table
        self.schema = schema
        self.ignore = ignore
        self.logger = logger
        if not isinstance(self.database, (pw.Database, pw.Proxy)):
            raise RuntimeError("Invalid database: %s" % database)

    @cached_property
    def model(self) -> typing.Type[MigrateHistory]:
        """Initialize and cache MigrationHistory model."""
        MigrateHistory._meta.database = self.database
        MigrateHistory._meta.table_name = self.migrate_table
        MigrateHistory._meta.schema = self.schema
        MigrateHistory.create_table(True)
        return MigrateHistory

    @property
    def todo(self):
        raise NotImplementedError

    @property
    def done(self):
        """Scan migrations in database."""
        return [mm.name for mm in self.model.select().order_by(self.model.id)]

    @property
    def diff(self):
        """Calculate difference between fs and db."""
        done = set(self.done)
        return [name for name in self.todo if name not in done]

    @cached_property
    def migrator(self):
        """Create migrator and setup it with fake migrations."""
        migrator = Migrator(self.database, self.schema)
        for name in self.done:
            self.run_one(name, migrator)
        return migrator

    def create(self, name="auto", auto=False):
        """Create a migration.
        :param auto: Python module path to scan for models.
        """
        migrate = rollback = ""
        if auto:
            # Need to append the CURDIR to the path for import to work.
            sys.path.append(CURDIR)
            try:
                modules = [auto]
                if isinstance(auto, bool):
                    modules = [m for _, m, ispkg in pkgutil.iter_modules([CURDIR]) if ispkg]

                models = [m for module in modules for m in load_models(module)]

            except ImportError:
                return self.logger.exception("Can't import models module")

            if self.ignore:
                models = [m for m in models if m._meta.name not in self.ignore]

            for migration in self.diff:
                self.run_one(migration, self.migrator)

            migrate = compile_migrations(self.migrator, models)
            if not migrate:
                return self.logger.warning("No changes found.")

            rollback = compile_migrations(self.migrator, models, reverse=True)

        self.logger.info('Creating migration "%s"', name)
        name = self.compile(name, migrate, rollback)
        self.logger.info('Migration has been created as "%s"', name)
        return name

    def merge(self, name="initial"):
        """Merge migrations into one."""
        migrator = Migrator(self.database)
        migrate = compile_migrations(migrator, self.migrator.state.values())
        if not migrate:
            return self.logger.error("Can't merge migrations")

        self.clear()

        self.logger.info('Merge migrations into "%s"', name)
        rollback = compile_migrations(self.migrator, [])
        name = self.compile(name, migrate, rollback, 0)

        migrator = Migrator(self.database)
        self.run_one(name, migrator, change_schema=False, change_history=True)
        self.logger.info('Migrations has been merged into "%s"', name)

    def clear(self):
        """Clear migrations."""
        self.model.delete().execute()

    def compile(self, name, migrate="", rollback="", num=None):
        raise NotImplementedError

    def read(self, name):
        raise NotImplementedError

    def run_one(
        self,
        name: str,
        migrator: Migrator,
        change_schema: bool = False,
        change_history: bool = False,
        downgrade: bool = False,
    ) -> None:
        """Run/emulate a migration with given name."""
        fake = not change_schema
        try:
            migration = self.read(name)

            def run_migrator():
                if not downgrade:
                    self.logger.info('Migrate "%s"', name)
                    migration.migrate(migrator, self.database, fake=fake)
                    migrator.run(change_schema)
                    if change_history:
                        self.model.create(name=name)
                else:
                    self.logger.info("Rolling back %s", name)
                    migration.rollback(migrator, self.database, fake=fake)
                    migrator.run(change_schema)
                    if change_history:
                        self.model.delete().where(self.model.name == name).execute()

                self.logger.info("Done %s", name)

            if migration.atomic:
                with self.database.transaction():
                    run_migrator()
            else:
                run_migrator()

        except Exception:
            operation = "Migration" if not downgrade else "Rollback"
            self.logger.exception("%s failed: %s", operation, name)
            raise

    def run(self, name=None, fake=False):
        """Run migrations."""
        self.logger.info("Starting migrations")

        done = []
        diff = self.diff
        if not diff:
            self.logger.info("There is nothing to migrate")
            return done

        migrator = self.migrator
        for mname in diff:
            self.run_one(mname, migrator, change_schema=not fake, change_history=True)
            done.append(mname)
            if name and name == mname:
                break

        return done

    def rollback(self, name):
        name = name.strip()
        done = self.done
        if not done:
            raise RuntimeError("No migrations are found.")
        if name != done[-1]:
            raise RuntimeError("Only last migration can be canceled.")

        migrator = self.migrator
        self.run_one(name, migrator, change_schema=True, downgrade=True, change_history=True)
        self.logger.warning("Downgraded migration: %s", name)


def make_ext_import(database: pw.Database) -> str:
    if isinstance(database, Psycopg3Database):
        return "import playhouse.psycopg3_ext as pw_pext"
    elif isinstance(database, pw.PostgresqlDatabase):
        return "import playhouse.postgres_ext as pw_pext"
    return ""


class Router(BaseRouter):
    filemask = re.compile(r"[\d]{3}_[^\.]+\.py$")

    def __init__(self, database, migrate_dir=DEFAULT_MIGRATE_DIR, **kwargs):
        super(Router, self).__init__(database, **kwargs)
        self.migrate_dir = migrate_dir

    @property
    def todo(self):
        """Scan migrations in file system."""
        if not os.path.exists(self.migrate_dir):
            self.logger.warning("Migration directory: %s does not exist.", self.migrate_dir)
            os.makedirs(self.migrate_dir)
        return sorted(f[:-3] for f in os.listdir(self.migrate_dir) if self.filemask.match(f))

    def compile(self, name, migrate="", rollback="", num=None):
        """Create a migration."""
        if num is None:
            num = len(self.todo)

        name = "{:03}_".format(num + 1) + name
        filename = name + ".py"
        path = os.path.join(self.migrate_dir, filename)
        with open(path, "w") as f:
            f.write(
                MIGRATE_TEMPLATE.format(
                    migrate=migrate, rollback=rollback, name=filename, ext_import=make_ext_import(self.database)
                )
            )

        return name

    def read(self, name):
        """Read migration from file."""
        call_params = {}
        if os.name == "nt" and sys.version_info >= (3, 0):
            # if system is windows - force utf-8 encoding
            call_params["encoding"] = "utf-8"
        with open(os.path.join(self.migrate_dir, name + ".py"), **call_params) as f:
            code = f.read()
            scope = {}
            exec_in(code, scope)

            atomic, migrate, rollback = (
                scope.get("__ATOMIC", True),
                scope.get("migrate", VOID),
                scope.get("rollback", VOID),
            )

            class _Migration(Migration):
                pass

            _Migration.atomic = atomic
            _Migration.migrate = migrate
            _Migration.rollback = rollback
            return _Migration

    def clear(self):
        """Remove migrations from fs."""
        super(Router, self).clear()
        for name in self.todo:
            filename = os.path.join(self.migrate_dir, name + ".py")
            os.remove(filename)


def load_models(module):
    """Load models from given module."""
    modules = _import_submodules(module)
    return {m for module in modules for m in filter(_check_model, (getattr(module, name) for name in dir(module)))}


def _import_submodules(package, passed=UNDEFINED):
    if passed is UNDEFINED:
        passed = set()

    if isinstance(package, str):
        package = import_module(package)

    modules = []

    for _loader, name, is_pkg in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        if name in passed:
            continue
        passed.add(name)

        module = sys.modules.get(name)
        if module is None:
            module = import_module(name)

        modules.append(module)
        if is_pkg:
            modules += _import_submodules(module, passed=passed)

    return modules


def _check_model(obj, models=None):
    """Checks object if it's a peewee model and unique."""
    return isinstance(obj, type) and issubclass(obj, pw.Model) and hasattr(obj, "_meta")


def compile_migrations(migrator, models, reverse=False):
    """Compile migrations for given models."""

    migrations = diff_many(models, migrator.state.values(), reverse=reverse)
    if not migrations:
        return False

    migrations = NEWLINE + NEWLINE.join("\n\n".join(migrations).split("\n"))
    return CLEAN_RE.sub("\n", migrations)



from __future__ import annotations

import hashlib
import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any

import peewee as pw

from miggy.types import ModelCls

if TYPE_CHECKING:
    from collections.abc import Sequence

    from miggy.types import ModelCls


def exec_in(code, glob, loc=None):
    if isinstance(code, str):
        code = compile(code, "<string>", "exec", dont_inherit=True)
    exec(code, glob, loc)


class Default(pw.ColumnBase):
    def __init__(self, value: str) -> None:
        self.value = value

    def __sql__(self, ctx) -> Any:
        ctx.literal("DEFAULT (%s)" % self.value)
        return ctx

    @classmethod
    def from_SQL(cls, sql: pw.SQL) -> None | Default:
        """
        Parse the constraint from raw sql
        """
        _sql = sql.sql.strip()
        if sql.params or not _sql.lower().startswith("default "):
            return None
        value = _sql.split(maxsplit=1)[1]
        if not value.strip():
            return None
        return Default(value=_sql.split(maxsplit=1)[1])


def get_default_constraint(field: pw.Field) -> None | Default:
    if field.constraints is None:
        return None
    constraints = []
    for constraint in field.constraints:
        if isinstance(constraint, Default):
            constraints.append(constraint)
        elif isinstance(constraint, pw.SQL):
            if _constraint := Default.from_SQL(constraint):
                constraints.append(_constraint)
    if len(constraints) > 1:
        raise ValueError(f'"{field.name}" field has more than one default constraint')
    return constraints[0] if constraints else None


def get_default_constraint_value(field: pw.Field):
    c = get_default_constraint(field)
    return c.value if c else None


def _truncate_constraint_name(constraint, maxlen=64):
    if len(constraint) > maxlen:
        name_hash = hashlib.md5(constraint.encode("utf-8")).hexdigest()
        constraint = "%s_%s" % (constraint[: (maxlen - 8)], name_hash[:7])
    return constraint


class ModelIndex(pw.ModelIndex):
    def __init__(
        self,
        model: ModelCls,
        fields: Sequence[pw.Field],
        unique: bool = False,
        safe: bool = True,
        where: pw.SQL | None = None,
        concurrently=False,
        using=None,
        name: str | None = None,
    ) -> None:
        self.concurrently = concurrently
        super().__init__(model=model, fields=fields, unique=unique, safe=safe, where=where, using=using, name=name)

    def _generate_name_from_fields(self, model, fields):
        accum = [field.column_name for field in fields]

        if not accum:
            raise ValueError("Unable to generate a name for the index, please explicitly specify a name.")

        clean_field_names = re.sub(r"[^\w]+", "", "_".join(accum))
        prefix = model._meta.table_name
        return _truncate_constraint_name("_".join((prefix, clean_field_names)))

    def __sql__(self, ctx):
        context = super().__sql__(ctx)
        if self.concurrently:
            context._sql.insert(1, "CONCURRENTLY ")
        return context


def has_single_index(field: pw.Field) -> bool:
    return field.index or field.unique


def make_single_index(field: pw.Field) -> ModelIndex:
    return ModelIndex(field.model, (field,), unique=field.unique, safe=False, using=field.index_type)


def get_single_index_name(field: pw.Field) -> str:
    return make_single_index(field)._name


def get_single_index(field: pw.Field) -> pw.Model:
    if has_single_index(field):
        return make_single_index(field)
    return None


def indexes_state(model_cls: pw.Model) -> dict[str, ModelIndex]:
    if not hasattr(model_cls._meta, "indexes_state"):
        model_cls._meta.indexes_state = {}
    return model_cls._meta.indexes_state


def copy_model(model_cls: ModelCls) -> ModelCls:
    # this function based on ModelBase.__new__ logic
    attrs = {}
    # copying fields
    for k, v in model_cls.__dict__.items():
        if isinstance(v, pw.FieldAccessor):
            attrs[k] = deepcopy(v.field)
    # copying Meta
    meta_options = {}
    if hasattr(model_cls, "_meta"):
        base_meta = model_cls._meta
        meta_keys = ["legacy_table_names", "table_name", "database", "indexes_state"]
        for k in meta_keys:
            try:
                meta_options[k] = base_meta.__dict__[k]
            except KeyError:
                pass
        attrs["Meta"] = type("Meta", (object,), meta_options)
    return type(model_cls.__name__, model_cls.__bases__, attrs)


def delete_field(model: ModelCls, field: pw.Field) -> None:
    """Delete field from model."""
    model._meta.remove_field(field.name)
    delattr(model, field.name)
    if isinstance(field, pw.ForeignKeyField):
        delattr(model, field.object_id_name)
        delattr(field.rel_model, field.backref)



import collections
import datetime as dt
import random
from collections.abc import Callable
from typing import Any

import peewee as pw
from playhouse.postgres_ext import BinaryJSONField, DateTimeTZField


class Counter(collections.Counter):
    def inc(self, element: str) -> int:
        self.update(**{element: 1})
        return self[element]


counter = Counter()


def char_enum_field_factory(field: pw.Field) -> str:
    first_choice = field.choices[0]
    first_value: str = first_choice[0]
    return first_value


def char_field_factory(field: pw.Field) -> str:
    value = counter.inc(field.name)
    return "%s%s" % (field.name, value)


def integer_field_factory(field: pw.Field) -> int:
    return counter.inc(field.name)


def json_field_factory(field: pw.Field) -> dict[str, Any]:
    value = counter.inc(field.name)
    return {f"key{value}": f"value{value}"}


field_type_map = {
    pw.DateField: lambda _: dt.date.today(),  # noqa: DTZ011
    pw.DateTimeField: lambda _: dt.datetime.now(),  # noqa: DTZ005
    DateTimeTZField: lambda _: dt.datetime.now(tz=dt.timezone.utc),
    pw.CharField: char_field_factory,
    pw.TextField: char_field_factory,
    pw.IntegerField: integer_field_factory,
    pw.SmallIntegerField: integer_field_factory,
    pw.BooleanField: lambda _: False,
    pw.BigIntegerField: lambda _: random.randint(1, 9999999),
    BinaryJSONField: json_field_factory,
}

_missing = object()


class FieldNotFound(Exception):
    pass


FieldMap = dict[type[pw.Field], Callable[[pw.Field], Any]]


def model_factory(
    model: type[pw.Model],
    custom_field_type_map: FieldMap | None = None,
    fill_nullable_values: bool = False,
    **kwargs: Any,
) -> pw.Model:
    """
    Create and save an instance of a Peewee model, automatically populating
    all required fields.

    Parameters
    ----------
    model : type[pw.Model]
        The Peewee model class to instantiate.
    custom_field_type_map : FieldMap | None, optional
        A mapping that allows customizing factory functions for specific
        field types. If ``None``, default handlers are used.
    fill_nullable_values : bool, optional
        Whether to automatically fill nullable fields as well.
    \*\*kwargs : Any
        ``field_name=value`` pairs that override or provide values for particular
        model fields.

    Returns
    -------
    pw.Model
        The newly created and saved model instance.

    Example
    -------
    >>> book = model_factory(Book, name="mytestname")
    """
    _field_type_map = field_type_map.copy()
    if custom_field_type_map is not None:
        _field_type_map.update(custom_field_type_map)

    nm = model()
    for field_name in kwargs:
        if field_name not in model._meta.fields:
            raise FieldNotFound(f'{model.__name__} has no "{field_name}" field')
    for field in model._meta.fields.values():
        field_name = field.name
        field_type = type(field)
        field_value = kwargs.get(field_name, _missing)
        if field_value is not _missing:
            pass
        elif field.primary_key or field.default is not None:
            continue
        elif field.null and not fill_nullable_values:
            field_value = None
        elif field_type is pw.ForeignKeyField:
            field_value = model_factory(
                field.rel_model, custom_field_type_map=_field_type_map, fill_nullable_values=fill_nullable_values
            )
        else:
            field_factory = _field_type_map[field_type]
            field_value = field_factory(field)
        setattr(nm, field_name, field_value)
    nm.save(force_insert=True)
    return nm
